import logging
import os

from cythonbuilder import appsettings
from cythonbuilder.definitions import PACKAGE_ROOT
from cythonbuilder.logs import create_logger, LoggerSettings


# Logger
# spacing = max([len(os.path.splitext(file)[0]) for file in os.listdir(PACKAGE_ROOT) if (file[:2] != "__")])
loggername = appsettings.package_name
loggerSettings = LoggerSettings(format=f'[%(name)s] %(message)s', level=logging.INFO)
loggerSettings.set_level_style(level='debug', color='green')
loggerSettings.set_level_style(level='info', color='white')
logger = create_logger(name=loggername, settings=loggerSettings)


# for lname, logger in logging.root.manager.loggerDict.items():
#     if (not isinstance(logger, logging.Logger)):
#         continue
#     if (loggername != logger):
#         logger.setLevel(level=logging.ERROR)



import os

import coloredlogs
from coloredlogs import ColoredFormatter

import logging
from logging.handlers import TimedRotatingFileHandler



# def DEFAULT_LOG_FORMAT(spacing:int=12):
#     return f'[%(name)s] %(asctime)s %(module)-{spacing}s %(lineno)-3d  %(message)s %(funcName)s'

DEFAULT_DATE_FORMAT = '%H:%M:%S'

DEFAULT_FIELD_STYLES = dict(
    asctime=dict(color='white'),
    hostname=dict(color='white'),
    levelname=dict(color='white', bold=True),
    name=dict(color='white'),
    programname=dict(color='white'),
    username=dict(color='white'),
    lineno=dict(color='white'),
    module=dict(color='white'),
    funcName=dict(color='white')
)

DEFAULT_LEVEL_STYLES = dict(
    # spam=dict(color='white', faint=True),
    # success=dict(color='green', bold=True),
    # notice=dict(color='magenta'),
    # verbose=dict(color='blue'),
    debug=dict(color='white'),
    info=dict(color='blue'),
    warning=dict(color='yellow', bright=True),
    error=dict(color='red', bold=True),
    critical=dict(color='black', bold=True, background='red'),
)
#endregion


class LoggerSettings:
    _format:str
    _dateformat:str
    _field_styles:dict
    _level_style:dict
    _log_file:str
    _level:int
    _valid_colors:[str] = ['black', 'blue', 'cyan', 'green', 'magenta', 'red', 'white', 'yellow']


    def __init__(
            self,
            level:int=logging.DEBUG,
            log_file:str=None,
            spacing:int=12,
            format:str=None,
            date_format:str=None,
    ):
        DEFAULT_FORMAT = f'[%(name)s] %(asctime)s %(module)-{spacing}s %(lineno)-3d  %(message)s'
        self._format = DEFAULT_FORMAT if (format == None) else format
        self._dateformat = DEFAULT_DATE_FORMAT if (date_format == None) else date_format
        self._field_styles = DEFAULT_FIELD_STYLES.copy()
        self._level_style = DEFAULT_LEVEL_STYLES.copy()
        self._level = level
        self._log_file = log_file


    def set_format(self, format:str):
        """" Sets the format of which fields will be taken """
        self._format = format

    def set_date_format(self, dateformat: str):
        """ Format the way asctime is displayed """

        self._dateformat = dateformat

    def set_level_style(self, level:str, color:str, background:str=None, bold:bool=False, bright:bool=False, faint:bool=False) -> None:
        """ Sets the styling of a certain level of the logger """

        level = level.lower()

        # Check inputs
        if (level == 'exception'):
            raise ValueError("Cannot set style of 'exception' directly since 'exception' is styled similar to 'error'. Set the style of 'error' instead")
        if (self._level_style.get(level) == None):
            raise ValueError(f"Level '{level}' does not exits. Existing levels: {list(self._level_style)}")

        # Cleanup
        color = color.lower()
        if (background != None):
            background = background.lower()

        # Set styles
        self._level_style[level] = {"color": color}
        if (background != None):
            self._level_style[level]['background'] = background
        if (bold != None):
            self._level_style[level]['bold'] = bold
        if (bright != None):
            self._level_style[level]['bright'] = bright
        if (faint != None):
            self._level_style[level]['faint'] = faint

    def set_field_styles(self, field:str, color:str, background:str=None, bold:bool=False, bright:bool=False, faint:bool=False) -> None:
        """ Sets the styling of fields in the logs """

        # Check inputs
        if (self._field_styles.get(field) == None):
            raise ValueError(f"Field '{field}' does not exits. Existing levels: {list(self._field_styles)}")

        # Cleanup
        color = color.lower()
        if (background != None):
            background = background.lower()

        # Set styles
        self._field_styles[field] = {"color": color}
        if (background != None):
            self._field_styles[field]['background'] = background
        if (bold != None):
            self._field_styles[field]['bold'] = bold
        if (bright != None):
            self._field_styles[field]['bright'] = bright
        if (faint != None):
            self._field_styles[field]['faint'] = faint


def create_logger(name:str, log_file:str=None, settings:LoggerSettings=None):
    """ Creates a logger """

    # 0. Check if logger with that name already exists
    if (name in list_all_loggers()):
        raise ValueError("a logger with this name already exists")

    # 1. Fall back to default settings if none passed
    if (settings == None):
        settings = LoggerSettings()

    # 2. Instantiate the logging module
    logging.basicConfig()
    logger = logging.getLogger(name=name)

    # 3. Set up a coloredformatter that determines the color styles and install it
    coloredFormatter = ColoredFormatter(
        fmt=settings._format,
        datefmt=settings._dateformat,
        level_styles=settings._level_style,
        field_styles=settings._field_styles
    )
    coloredlogs.install(
        level=settings._level,
        logger=logger,
        fmt=settings._format,
        datefmt=settings._dateformat
    )
    logger.propagate = False  # prevents logs to pass to the root logger (needs to be after coloredlogs.install)

    # 4. Set up handlers for logfile and the console
    if (log_file != None):
        fh = TimedRotatingFileHandler(filename=log_file, when='D', interval=1, backupCount=90, encoding='utf-8', delay=False)
        fh.setFormatter(fmt=coloredFormatter)
        logger.addHandler(hdlr=fh)
    # Console handler
    ch = logging.StreamHandler()
    ch.setFormatter(fmt=coloredFormatter)
    logger.addHandler(hdlr=ch)    # def set_

    return logger

def list_all_loggers():
    """ Return the names of all active loggers """
    return list(logging.root.manager.loggerDict)

def set_only_active_loggers(active_logger_names:[str], level:int, other_loggers_level:int):
    """ Only these loggers will be set to 'level'. All other loggers will be set to 'other_loggers_level' """
    for loggername in list_all_loggers():
        if (loggername in active_logger_names):
            set_level(loggername=loggername, level=level)
        else:
            set_level(loggername=loggername, level=other_loggers_level)

def set_format(logger:logging.Logger, format:str):
    """ Sets the level of a logger """
    handler:logging.Handler
    for handler in logger.handlers:

        formatter:ColoredFormatter = handler.formatter
        coloredFormatter = ColoredFormatter(
            fmt=format,
            datefmt=formatter.datefmt,
            level_styles=formatter.level_styles,
            field_styles=formatter.field_styles
        )
        handler.setFormatter(fmt=coloredFormatter)

def set_logger_debug_mode(logger:logging.Logger):
    """ Set the debug mode of the logger """

    set_format(logger=logger, format=f"[%(name)s] %(asctime)s %(module)-8s %(lineno)-3d  %(message)s")
    logger.setLevel(logging.DEBUG)



import sys

from cythonbuilder import cython_builder
from cythonbuilder import appsettings
from cythonbuilder import helpers
from cythonbuilder import set_logger_debug_mode
from cythonbuilder.services import logger

def display_help():
    print(f"""{appsettings.package_name}
        Automatically builds and packages your Cython code 
        1. Initialize {appsettings.package_name} with `{appsettings.console_name} init` 
        3. Call `{appsettings.console_name} build` to build and package all .pyx files in your project 

        Commands:
        (call either command with -v to display more information)
        help        Show this screen
        init        Create 'ext' and 'ext/annotations' folders
        list        Lists all pyx files in your project that can be built
                    Pass a space-separated list of filenames (with or without .pyx) to filter
        build       Finds and builds all pyx files in your project (excluding venv directories) package cython files. Then cleans the project (see cybuilder clean)
                    Pass a space-separated list of filenames (with or without .pyx) to filter.
                    Options:
                      --no-interfaces   Prevents creating .pyi interface files (default False)
                      --include-numpy   Includes numpy (default False)
                      --no-annotation   Disables generating the annotations html (default True)
                      --no-cleanup      Prevents removal of intermediate C files generated by Cython (default True)
        clean       Moves all generated .html files to ext/annotations 
                    Pass a space-separated list of filenames (with or without .pyx) to filter.
                    Options:
                      --keep-c      Prevents removal of intermediate C files generated by Cython (default False)
        interface   Generates all pyi files that serve as an interface for IDE's
                    Pass a space-separated list of filenames (with or without .pyx) to filter.
    """)
    sys.exit(0)


def main():
    """ Parse command line """

    args = sys.argv[1:]

    # Preferences
    # FORCE = '-f' in args or '--force' in args
    VERBOSE = '-v' in args or '--verbose' in args
    ACCEPT = '-y' in args or '--yes' in args
    # OVERWRITE = '-o' in args or '--overwrite' in args
    for flag in ['-f', '--force', '-v', '--verbose', '-y', '--accept', '-o', '--overwrite']:
        args.remove(flag) if (flag in args) else None


    # Logging level
    if (VERBOSE):
        set_logger_debug_mode(logger=logger)
        # from cythonbuilder.logs import set_format
        # set_format(logger=logger, format=f"[%(name)s] %(asctime)s %(module)-8s %(lineno)-3d  %(message)s")
        # logger.setLevel(logging.DEBUG)



    # Parse arguments - Must provide at least one argument
    if (len(args) == 0):
        display_help()
    cmd1 = helpers.CliTools.pop_arg_or_exit(arglist=args, errormessage=f"[{appsettings.package_name}] {appsettings.package_name} expects at least one argument. Check out [helpy help] for more information")
    logger.debug(msg=f"{cmd1=}")


    # HELPY functions
    if (cmd1 == 'init'):
        logger.info(msg=f"[{cmd1}] - Initializing..")
        cython_builder.cy_init()
        logger.info(msg=f"[{cmd1}] - Initialized")
        sys.exit(0)
    elif (cmd1 == 'list'):
        # Lists all pyx files that can be built
        target_filenames = args if (len(args) > 0) else None
        filtermsg = f" (filtered by {' '.join(target_filenames)}) " if (target_filenames) else ""
        logger.info(msg=f"Listing all pyx files{filtermsg}..")
        found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
        if (len(found_pyx_files) >0):
            found_files_string = "\n".join([f"\t - {fle}" for fle in found_pyx_files])
            logger.info(msg=f"Found files:\n{found_files_string}")
        else:
            logger.info(msg=f"No pyx files found")
    elif (cmd1 == 'build'):
        # Get arguments
        include_numpy = '--include-numpy' in args
        dont_generate_annotations = '--no-annotations' in args
        dont_generate_pyi = '--no-interfaces' in args
        keep_c_files = '--no-cleanup' in args
        args = [a for a in args if (a not in ['--no-annotations', '--include-numpy', '--no-cleanup', '--no-interfaces'])]
        target_filenames = args if (len(args) > 0) else None

        # Validate arguments
        numpy_is_installed = helpers.package_is_installed(package_import_name='numpy')
        if (include_numpy and not numpy_is_installed):
            if (input(f"[{appsettings.package_name} {cmd1}] You want to include numpy but your current project does not have numpy installed. Install numpy? (y/n)").lower() == 'y'):
                logger.warning(msg=f"Please install numpy with [pip install numpy]")
                sys.exit(0)
            else:
                include_numpy = False
        if (not include_numpy and numpy_is_installed):
            if (input(f"[{appsettings.package_name} {cmd1}] Your current project uses numpy. Do you want to include numpy in your Cython build? (y/n)").lower() == 'y'):
                include_numpy = True

        logger.info(msg=f"[{cmd1}] - Building Cython files..")
        try:
            # 1. Find pyx files
            found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
            if (len(found_pyx_files) == 0):
                logger.info(msg="No .pyx files found")
                sys.exit(0)
            logger.debug(msg=f"Found {len(found_pyx_files)} to build")

            # 2. Confirm that we want to build
            if (not ACCEPT):
                __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
                if (input(f"[{appsettings.package_name} {cmd1}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                    logger.info(msg="Exiting..", )
                    sys.exit(0)
            logger.debug(msg=f"Building {len(found_pyx_files)} pyx files..")

            # 3. Build
            cython_builder.cy_build(
                target_files=found_pyx_files,
                create_annotations=not dont_generate_annotations,
                include_numpy=include_numpy,
            )
            logger.debug(msg=f"Built {len(found_pyx_files)} pyx files, cleaning up..")

            # 4. Cleanup after build
            cython_builder.cy_clean(target_files=found_pyx_files, keep_c_files=keep_c_files)
            logger.debug(msg=f"Cleanup complete")

            # 5.  Generate pyi files
            if (not dont_generate_pyi):
                logger.debug(msg=f"Generating interface files..")
                cython_builder.cy_interface(target_files=found_pyx_files)
                logger.debug(msg=f"Generated pyi files")

            logger.info(msg=f"[{cmd1}] - Cython build success")
        except Exception as e:
            logger.error(msg=f"[{cmd1}] build error: {e}")
            sys.exit(1)
    elif (cmd1 == 'clean'):
        # Get arguments
        keep_c_files = '--keep-c' in args
        args = [a for a in args if (a not in ['--keep-c'])]
        target_filenames = args if (len(args) > 0) else None

        # Clean
        logger.info(msg=f"[{cmd1}] - Cleaning Cython files..")
        try:
            # 1. Find pyx files
            found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
            if (len(found_pyx_files) == 0):
                logger.info(msg="No .pyx files found")
                sys.exit(0)
            logger.debug(msg=f"Found {len(found_pyx_files)} to clean")

            # 2. Confirm that we want to build
            if (not ACCEPT):
                __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
                if (input(f"[{appsettings.package_name} {cmd1}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                    logger.info(msg="Exiting..", )
                    sys.exit(0)
            logger.debug(msg=f"Cleaning {len(found_pyx_files)} pyx files..")

            # 3. Clean
            cython_builder.cy_clean(target_files=found_pyx_files, keep_c_files=keep_c_files)
            logger.debug(msg=f"Cleanup complete")
            logger.info(msg=f"[{cmd1}] - Cython build success")
        except Exception as e:
            logger.error(msg=f"[{cmd1}] build error: {e}")
            sys.exit(1)
    elif (cmd1 in ['interface', 'pyi']):

        target_filenames = args if (len(args) > 0) else None

        logger.info(msg=f"[{cmd1}] - Generating interface files..")
        try:
            # 1. Find pyx files
            found_pyx_files: [str] = cython_builder.cy_list(target_files=target_filenames)
            if (len(found_pyx_files) == 0):
                logger.info(msg="No .pyx files found")
                sys.exit(0)
            logger.debug(msg=f"Found {len(found_pyx_files)} to create pyi files for")

            # 2. Confirm that we want to build
            if (not ACCEPT):
                __formatted_package_list = "\n".join(f"\t - {file}" for file in found_pyx_files)
                if (input(f"[{appsettings.package_name} {cmd1}] these {len(found_pyx_files)} pyx files?\n(y/n) \n {__formatted_package_list}").lower() != "y"):
                    logger.info(msg="Exiting..", )
                    sys.exit(0)
            logger.debug(msg=f"Creating {len(found_pyx_files)} pyi files..")

            # 3. Generate pyi files
            logger.debug(msg=f"Generating pyi files..")
            cython_builder.cy_interface(target_files=found_pyx_files)
            logger.debug(msg=f"Generated pyi files")
            logger.info(msg=f"[{cmd1}] - Generating interface files success")
        except Exception as e:
            logger.error(msg=f"[{cmd1}] build error: {e}")
            sys.exit(1)

    elif (cmd1 == 'help'):
        display_help()
    else:
        logger.warning(msg=f"unknown command: '{cmd1}'")
        display_help()




import os
import sys

from cythonbuilder import appsettings

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))
PACKAGE_ROOT = sys.modules[appsettings.package_name].__path__[0]




import os
import shutil

from .definitions import PACKAGE_ROOT
from .services import logger


class FilesAndFolders:

    @staticmethod
    def create_folder(folderpath: str) -> None:
        if (os.path.exists(folderpath)):
            logger.debug(msg=f"{FilesAndFolders.create_folder.__name__}] - provided folder already exists: '{folderpath}', skipping..")
            return
        os.mkdir(path=folderpath)
        logger.debug(msg=f"{FilesAndFolders.create_folder.__name__}] - created folder")

    @staticmethod
    def remove_folder(folderpath: str) -> None:
        if (not os.path.isdir(folderpath)):
            logger.debug(msg=f"{FilesAndFolders.remove_folder.__name__}] - provided folder cannot be removed: it does not exist: '{folderpath}', skipping..")
            return
        shutil.rmtree(folderpath)
        logger.debug(msg=f"{FilesAndFolders.remove_folder.__name__}] - Removed folder {folderpath}")

    @staticmethod
    def create_empty_file(filepath: str, overwrite: bool = False) -> None:
        """ Creates an empty file if not exists """
        if (not overwrite):
            if (os.path.isfile(filepath)):
                logger.debug(msg=f"{FilesAndFolders.create_empty_file.__name__}] - provided filepath already exists '{filepath}', skipping..")
                return
        with open(filepath, 'w') as file:
            file.write("")

    @staticmethod
    def replace_in_file(targetfilepath: str, replace_this_text: str, replacment_text: str) -> None:
        """ Replaces a text with the replacement text in a text file """

        # Validate
        if (not os.path.isfile(targetfilepath)):
            raise ValueError(f"{FilesAndFolders.replace_in_file.__name__}] - provided filepath invalid; does not exist: {targetfilepath}")

        # Open target, replace string and write again
        with open(targetfilepath, 'r') as file:
            filedata = file.read()
        filedata = filedata.replace(replace_this_text, replacment_text)
        with open(targetfilepath, 'w') as file:
            file.write(filedata)

    @staticmethod
    def copy_file(targetfilename: str, destinationpath:str, overwrite: bool = False) -> None:
        """ download a file from a url to the given file path """

        source = os.path.join(PACKAGE_ROOT, 'files', targetfilename)
        if (not os.path.isfile(source)):
            raise ValueError(f"{FilesAndFolders.copy_file.__name__}] - provided filename invalid; source file does not exist: {source}")
        if (not os.path.isdir(os.path.dirname(destinationpath))):
            raise ValueError(f"{FilesAndFolders.copy_file.__name__}] - provided destination invalid; target folder does not exist: {os.path.dirname(destinationpath)}")

        # Overwrite?
        if (not overwrite):
            if (os.path.isfile(destinationpath)):
                logger.debug(msg=f"[{FilesAndFolders.copy_file.__name__}] - File {targetfilename} already exists. Skipping..")
                return

        # Copy
        logger.debug(f"copying {source} to {destinationpath}")
        with open(destinationpath, 'w') as dst:
            with open(source, 'r') as src:
                dst.write(src.read())

    @staticmethod
    def remove_file(targetfilename:str) -> None:
        """ Remove specified file """
        if (not os.path.isfile(targetfilename)):
            raise ValueError(f"{FilesAndFolders.remove_file.__name__}] - provided filename invalid; target file does not exist: {targetfilename}")
        os.remove(targetfilename)

    @staticmethod
    def move_file(srcfilename: str, dstfilename:str, overwrite: bool = False) -> None:
        """ download a file from a url to the given file path """

        if (not os.path.isfile(srcfilename)):
            raise ValueError(f"{FilesAndFolders.move_file.__name__}] - provided source filename invalid; source file does not exist: {srcfilename}")
        if (not os.path.isdir(os.path.dirname(dstfilename))):
            raise ValueError(f"{FilesAndFolders.move_file.__name__}] - provided destination folder invalid; does not exist: {os.path.isdir(os.path.dirname(dstfilename))}")

        if (os.path.isfile(dstfilename) and not overwrite):
            raise ValueError(f"{FilesAndFolders.move_file.__name__}] - cannot move file; target file already exists: {os.path.isdir(os.path.dirname(dstfilename))}")

        # Copy
        logger.debug(f"moving {srcfilename} to {dstfilename}")
        shutil.move(src=srcfilename, dst=dstfilename)


class CliTools:
    @staticmethod
    def pop_arg_or_exit(arglist: [str], errormessage: str):
        """ Tries to pop an arg from the list. If this is not possible: display errormessage and exit """
        if (len(arglist) <= 0):
            logger.error(msg=errormessage)
            quit()
        return arglist.pop(0).lower()


def package_is_installed(package_import_name:str=None) -> bool:
    """ Returns t/f depending on whether a package is installed in this project
        :arg package_import_name   str     name of the package you're checking
    """

    if (package_import_name == None or package_import_name == 'venv'):
        return False
    _package_installed = False
    try:
        exec(f"import {package_import_name}")
        _package_installed = True
    except ImportError:
        pass
    except Exception as e:
        raise e
    return _package_installed



from pathlib import Path
import os
import shutil
import sys

from cythonbuilder.helpers import FilesAndFolders
from cythonbuilder.services import logger
from cythonbuilder import appsettings
from cythonbuilder import pyigenerator

project_dir = os.getcwd()


def cy_init():
    """ Creates projdir.ext, projdir.ext.annotations """

    # create root.ext, root.ext.annotations
    FilesAndFolders.create_folder(folderpath=os.path.join(project_dir, appsettings.cython_extensions_dirname))
    FilesAndFolders.create_folder(folderpath=os.path.join(project_dir, appsettings.cython_anno_dirname))
    logger.debug(msg=f"[{cy_init.__name__}] - Initialized cybuilder at {project_dir}")



def cy_list(target_files:[str]=None) -> [str]:
    """ Target files is optional filter. Returns a list of fullpaths to pyxfiles """
    # 1. Find all fullpaths to pyx files in all folders but the /venv
    venv_paths_folders = [os.path.dirname(path) for path in Path(project_dir).rglob('pyvenv.cfg')]
    pyx_fullpaths = [str(pyxpath) for pyxpath in Path(project_dir).rglob('*.pyx')]
    for venvpath in venv_paths_folders:
        pyx_fullpaths = [pf for pf in pyx_fullpaths if (venvpath not in pf)]

    # 2. Apply optional file name filter
    if (target_files != None):
        my_pyx_file_names = [os.path.splitext(os.path.basename(p))[0] for p in pyx_fullpaths]
        target_file_names = [os.path.splitext(p)[0] for p in target_files]

        # Calculate overlap and difference
        diff = set(target_file_names).difference(set(my_pyx_file_names))
        intersection = set(my_pyx_file_names).intersection(set(target_file_names))

        # if (len(diff) > 0):
        #     raise ValueError(f"Cannot find provided target files: {[f'{f}.pyx' for f in diff]}")

        _pyx_fullpaths = []
        for pyx in pyx_fullpaths:
            for ins in intersection:
                if (f"{ins}.pyx" in pyx):
                    _pyx_fullpaths.append(pyx)
        pyx_fullpaths = _pyx_fullpaths
    return pyx_fullpaths
def cy_build(target_files:[str] = None, create_annotations:bool=True, include_numpy:bool=False):
    """ Builds all pyx files in the /ext folder """

    # 1. Get target files
    if (target_files == None):
        target_files = cy_list()


    # We want to build in place
    sys.argv = [sys.argv[0], 'build_ext', '--inplace']

    from setuptools import setup, Extension
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize
    import Cython.Compiler.Options

    # Annotation is whether or not the html should be created
    Cython.Compiler.Options.annotate = create_annotations

    # Create Extension objects
    ext_modules = []
    for n in target_files:
        if (not os.path.isfile(n)):
            logger.info(msg=f"File '{n}' not found; skipping..")
            continue
        logger.debug(msg=f"C {n}")
        module_name, extension = os.path.splitext(os.path.basename(n))
        # The name must be plain, no path
        obj = Extension(
            name=module_name,
            sources=[n],
            # extra_compile_args=["-O2", "-march=native"]
        )
        ext_modules.append(obj)

    # Extra include folders. Mainly for numpy.
    include_dirs = []
    if (include_numpy):
        try:
            import numpy
            include_dirs += [numpy.get_include()]
        except:
            raise ValueError('Numpy is required, but not found. Please install')

    # build
    setup(
        cmdclass={'build_ext': build_ext},
        include_dirs=include_dirs,
        ext_modules=cythonize(ext_modules),
        # buid_dir=path_build_dir
    )

def cy_clean(target_files:[str] = None, keep_c_files:bool=False):
    """ Clean up all files """
    logger.debug(msg=f"[{cy_clean.__name__}] - start cy_clean with {target_files}")

    # 1. Get target files
    if (target_files == None):
        target_files = cy_list()

    # Make sure cybuilder is init because we need to move files to /ext/annotations
    cy_init()

    annotations_dir = os.path.join(project_dir, appsettings.cython_anno_dirname)

    # Remove build folder
    build_folder_path = os.path.join(project_dir, 'build')
    if (os.path.isdir(build_folder_path)):
        shutil.rmtree(path=build_folder_path)

    for built_file in target_files:
        if (not os.path.isfile(built_file)):
            logger.info(msg=f"File {built_file} not found; skipping..")
            continue

        _filename = os.path.splitext(os.path.basename(built_file))[0]    # no ext

        # Clean up C files
        if (not keep_c_files):
            filepath = f"{os.path.splitext(built_file)[0]}.c"
            if (not os.path.isfile(filepath)):
                logger.debug(msg=f"[{cy_clean.__name__}]: cannot remove file: does not exist {filepath}")
                continue

            FilesAndFolders.remove_file(targetfilename=f"{os.path.splitext(built_file)[0]}.c")
        # Move annotation html files
        src_htmlpath = f"{os.path.splitext(built_file)[0]}.html"
        dst_htmlpath = os.path.join(annotations_dir, os.path.basename(src_htmlpath))
        logger.debug(msg=f"[{cy_clean.__name__}] - Moving annotation files from {src_htmlpath} to {dst_htmlpath}")
        FilesAndFolders.move_file(
            srcfilename=src_htmlpath,
            dstfilename=dst_htmlpath,
            overwrite=True
        )
        # Move PYD files
        built_file_folder = os.path.dirname(built_file)
        pyd_file = ([fn for fn in os.listdir(project_dir) if (_filename in fn and ('.pyd' in fn or '.so' in fn))])[0]
        logger.debug(msg=f"[{cy_clean.__name__}] - Moving pyd files from {os.path.join(project_dir, pyd_file)} to {os.path.join(built_file_folder, pyd_file)}")
        FilesAndFolders.move_file(
            srcfilename=os.path.join(project_dir, pyd_file),
            dstfilename=os.path.join(built_file_folder, pyd_file),
            overwrite=True
        )
def cy_interface(target_files:[str] = None):
    """ Creates .pyi interface files from the provided target_files """


    # 1. Get target files
    if (target_files == None):
        target_files = cy_list()
    logger.debug(msg=f"[{cy_interface.__name__}] - generating interface files for {len(target_files)} found pyx files")


    for pyx_fullpath in target_files:
        if (not os.path.isfile(pyx_fullpath)):
            logger.info(msg=f"File {pyx_fullpath} not found; skipping..")
            continue
        pyi_fullpath = f"{os.path.splitext(pyx_fullpath)[0]}.pyi"
        logger.debug(msg=f"Creating .pyi for {pyx_fullpath}")

        pyigenerator.read_write_pyx_to_pyi(target_pyx_path=pyx_fullpath, target_pyi_path=pyi_fullpath)




# import code from your package here to make them available outside the module like:
from cythonbuilder.cython_builder import cy_init, cy_clean, cy_build, cy_list, cy_interface
from cythonbuilder.services import logger
from cythonbuilder.logs import set_logger_debug_mode


__version__ = "0.1.10"
__last_build_datetime__ ="2022-05-16 17:00:09"
__last_publish_version__ ="0.1.9"
__last_publish_datetime__ ="2022-05-16 16:22:20"



import setuptools

__package_name__ = "cythonbuilder"
__version__ = "0.1.10"
__author__ = 'Mike Huls'
__license__ = "MIT"
__maintainer__ = "Mike Huls"
__email__ = "mikehuls42@gmail.com"
__status__ = "Development"
__description__ = "CythonBuilder; automated compiling and packaging of Cython code"
__project_url__ = "https://github.com/mike-huls/cythonbuilder"


# Reads the content of your README.md into a variable to be used in the setup below
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name=__package_name__,
    packages=[__package_name__],
    # For including non-python files (in this example all files in the ./files folder)
    # package_data={'': ['files/*']},
    version=__version__,
    license=__license__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=__author__,
    author_email=__email__,
    url=__project_url__,
    project_urls={
        "Source": "https://github.com/mike-huls/cythonbuilder/",
        "Bug Tracker": "https://github.com/mike-huls/cythonbuilder/issues",
        "Documentation": "https://github.com/mike-huls/cythonbuilder/blob/main/README.md/",
    },
    # Entry points make CLI access possible
    entry_points={
        'console_scripts': [
            f'cythonbuilder=cythonbuilder.cli:main',
            f'cybuilder=cythonbuilder.cli:main',
        ],
    },
    # Add packages that need to be installed along this package:
    install_requires=['Cython', 'coloredlogs'],
    python_requires='>=3',
    # Describe this package in a few keywords:
    keywords=["pypi", "Cython", "setup", "packaging", "compilation"],  # descriptive meta-data
    # https://pypi.org/classifiers
    classifiers=[                                   
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ]
)


import os

package_name = 'cythonbuilder'
console_name = 'cybuilder'
cython_build_dirname = 'build'
cython_extensions_dirname = 'ext'
cython_anno_dirname = os.path.join(cython_extensions_dirname, 'annotations')




import re


# from cythonbuilder import logger


def read_write_pyx_to_pyi(target_pyx_path: str, target_pyi_path:str):
    """ Reads all content from a pyx file, convers and writes a pyi """

    # 1. Read pyx file
    pyx_lines: [str] = []
    with open(target_pyx_path, 'r') as open_pyx:
        pyx_lines.extend(open_pyx.readlines())

    # 2. Skip empty lines
    pyx_lines = [l.strip("\n") for l in pyx_lines]
    pyx_lines = [l for l in pyx_lines if (len(l) > 0)]

    # 2. Determine file indentation
    all_indentations = [__get_line_indentation_spacecount(line=l) for l in pyx_lines]
    all_indentations = [i for i in all_indentations if (i > 0)]
    spaces_for_one_tab = min(all_indentations)
    # Extra check: test wheter all are divisible by the space_for_one_tab
    for indent in all_indentations:
        if (indent % spaces_for_one_tab != 0):
            raise ValueError(f"Found invalid indentation: {indent} not divisible by {spaces_for_one_tab}")



    # Convert lines to py
    py_lines:[str] = []
    for line_idx in range(len(pyx_lines)):
        # pyx line to filter
        pyx_line = pyx_lines[line_idx]

        # Line details
        cur_line_is_import = line_is_import(line=pyx_line)
        cur_line_is_function = line_is_function(line=pyx_line)
        cur_line_is_class = line_is_class(line=pyx_line)
        cur_line_spaces = __get_line_indentation_spacecount(line=pyx_lines[line_idx])
        cur_line_indentation: int = int(cur_line_spaces / spaces_for_one_tab) if (cur_line_spaces > 0) else cur_line_spaces
        cur_line_is_attribute = cur_line_indentation == 1 and not cur_line_is_function and not cur_line_is_class and not cur_line_is_import

        # Add python decorators like @property and @dataclass
        if (pyx_line.strip()[0] == '@'):
            py_lines.append(pyx_line)

        # FILTERING     Skip if indentation > 2. That way we are one indentation deeper than method content
        if (not cur_line_is_import and not cur_line_is_class and not cur_line_is_function and not cur_line_is_attribute):
            continue

        # FILTERING     Skip if previous line is a function def
        if (line_idx > 0):
            pyx_line_prev = pyx_lines[line_idx - 1]
            pyx_line_prev.strip(' ')
            if (line_is_function(line=pyx_line_prev)):
                pyx_line = f"{cur_line_spaces * ' '}..."


        # only keep lines when the next line has a greater indentation
        # if (line_idx == len(pyx_lines) - 1):
        #     print("breaking")
        #     break
        # current_indentation = __get_line_indentation_spacecount(line=pyx_lines[line_idx])
        # nextline_indentation = __get_line_indentation_spacecount(line=pyx_lines[line_idx + 1])
        # if (nextline_indentation <= current_indentation):
        #     continue


        py_line = pyx_line_to_pyi(line=pyx_line.strip("\n"), spaces_for_one_tab=spaces_for_one_tab)
        if (py_line != None):
            py_lines.append(py_line)


    # Post-Process
    post_process(py_lines=py_lines, spaces_for_one_tab=spaces_for_one_tab)

    with open(target_pyi_path, 'w') as pyi_out:
        pyi_out.writelines([f"{line}\n" for line in py_lines])
        # for line in py_lines:
        #     print(line+'\n')
        #     pyi_out.write(line)

def post_process(py_lines:[str], spaces_for_one_tab:int):
    """ """
    for i, line in enumerate(py_lines):
        words_array = line.split(" ")
        if (words_array[0] == 'class'):
            next_line = py_lines[i+1]
            if (next_line[:spaces_for_one_tab] != '    '):
                py_lines[i] = line + f"\n{' ' * spaces_for_one_tab}..."
    return py_lines
def line_is_import(line:str) -> bool:
    """ """
    line = line.strip()
    word_array = line.split(" ")
    first_word_is_import = word_array[0] in ['import', 'cimport']
    line_contains_from_import = word_array[0] == 'from' and 'import' in word_array
    return any([first_word_is_import, line_contains_from_import])
def line_is_class(line:str) -> bool:
    """ """
    return line[-1] == ":" and "class" in line.split(" ")
def line_is_function(line:str) -> bool:
    """ """
    line = line.strip(' ')
    endswith_colon = line[-1] == ':'
    return endswith_colon and not line_is_class(line=line)
def line_is_c_function(line:str) -> bool:
    is_c_function = False
    if (line_is_function(line=line)):
        line_parts_set = set(line.split(" "))
        cfunction_defs = {'cpdef', 'cdef'}
        if (len(line_parts_set.intersection(cfunction_defs)) > 0):
            is_c_function = True
    return is_c_function



def pyx_line_to_pyi(line: str, spaces_for_one_tab: int):
    """ Interprets a line and converts pyx to pyi
        :arg spaces_for_one_tab       how many spaces equals one level of indentation
    """


    # 1. Calculate indentation level
    spacecount = __get_line_indentation_spacecount(line=line)
    indentation:int = int(spacecount / spaces_for_one_tab) if (spacecount > 0) else spacecount
    spaces = " " * spaces_for_one_tab * (indentation + 1)


    # 2. Remove unrequired characters
    line = line.strip(' ')



    # 3. Keep only lines that start with def, cpdef or cdef
    begin_words = ['def', 'cdef', 'cpdef', 'class', 'cimport', 'import', 'from']
    if not any([line[:len(w)] == w for w in begin_words]):
        # todo this correct?
        if (indentation > 1):
            return


    # If class is in the line: make sure it adheres to python def
    if (" " in line):
        splitline = line.split(" ")
        if (splitline[1] == 'class'):
            line = " ".join(splitline[1:])
        if (splitline[0] == 'cimport' or splitline[0] == 'import'):
            splitline[0] = 'import'
            line = " ".join(splitline)

    if ('array' in line):
        print(line, indentation)


    # 4. Start replacing lines
    line_parts = line.split(" ")
    for word in ['cpdef', 'cdef']:
        if word in line_parts:
            word_idx = line_parts.index(word)
            line_parts[word_idx] = 'def'
    line = " ".join(line_parts)


    # 5. Handle function types
    if (line_is_function(line=line)):
        if ('->' in line):
            line = f"{line.split('->')[0].strip()}:"

        array_between_brackets = re.findall('\(.*?\)',line)
        for brackets in array_between_brackets:
            # Remove return type definition
            func_part_one_old = line.replace(f"{brackets}:", "")
            func_part_one_new = " ".join([func_part_one_old.split(" ")[0], func_part_one_old.split(" ")[-1]])
            line = line.replace(func_part_one_old, func_part_one_new)
            for argument in brackets.split(","):

                # Strip away parentheses and spaces
                argument = argument.strip("() ")

                # C types are split like 'int age'
                arg_array = [a for a in argument.split(" ") if (len(a) > 0)]
                if (len(arg_array) <= 1):
                    continue
                var_ctype = arg_array[0].strip(":")
                var_name = arg_array[1].strip(":")
                if (line_is_c_function(line=line)):
                    var_name, var_ctype = var_ctype, var_name
                py_type = convert_type_cy_to_py(cy_type=var_ctype)
                line = line.replace(argument, f'{var_name}:{py_type}')

        line = line + f"\n{spaces}..."
    return indentation * spaces_for_one_tab * ' ' + line

def convert_type_cy_to_py(cy_type:str):
    """ """
    # todo continue https://stackoverflow.com/questions/55451545/what-are-all-the-types-available-in-cython
    cy_type = cy_type.lower()

    if (cy_type in ['bint', 'bool']):
        return 'bool'
    elif (cy_type in ['char', 'short', 'int', 'long', 'long long']):
        return 'int'
    elif (cy_type in ['float', 'double', 'long double']):
        return 'float'
    elif (cy_type in ['float complex', 'double complex', 'complex']):
        return 'complex'
    elif (cy_type in ['char*', 'std::string', 'str']):
        return 'str'
    # print(f"{indentation * spaces_for_one_tab * ' '}{line}")


def __get_line_indentation_spacecount(line: str) -> int:
    """ How many spaces is the provided line indented? """
    spaces: int = 0
    if (len(line) > 0):
        if (line[0] == ' '):
            for letter_idx in range(len(line)):
                if (line[letter_idx + 1] != ' '):
                    spaces = letter_idx + 1
                    break
    return spaces


import unittest
# from dotenv import load_dotenv
# from config.definitions import ROOT_DIR

class TestExample(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # load_dotenv(dotenv_path=os.path.join(ROOT_DIR, 'config', 'conf', '.env'))

    def test_isEqual(self):
        """ Test example """

        # 1. Arrange
        def function_from_project(a:float, b:float):
            return a * b

        # 2. Act
        result = function_from_project(5, 5)

        # 3. Assert
        self.assertEqual(result, 25)

if __name__ == '__main__':
    unittest.main()
